/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer.datastructure;

import java.util.Arrays;
import java.util.List;

import de.hft.stuttgart.citygml.viewer.math.BoundingBoxCalculator;
import de.hft.stuttgart.citygml.viewer.math.Vector3d;
import de.hft.stuttgart.citygml.viewer.parser.FeatureMapper;

/**
 * An axis aligned bounding box represented by its two corners
 * 
 * @author Matthias Betz
 *
 */
public class BoundingBox {

	private Vector3d[] bbox;

	/**
	 * Creates an axis aligned bounding box containing all points of all polygons
	 * 
	 * @param polygons containing the points from which the box will be created
	 * @return the bounding box around all points
	 */
	public static BoundingBox of(List<Polygon> polygons) {
		return BoundingBoxCalculator.calculateBoundingBox(polygons);
	}
	
	public static BoundingBox ofPoints(List<? extends Vector3d> points) {
		return BoundingBoxCalculator.calculateBoundingBoxFromPoints(points);
	}

	/**
	 * Creates a new bounding box with two vectors as corner points.
	 * 
	 * @param box the array of length 2 containing both corners
	 * @return the new bounding box
	 */
	public static BoundingBox of(Vector3d[] box) {
		return new BoundingBox(box);
	}

	private BoundingBox(Vector3d[] bbox) {
		if (bbox == null || bbox.length != 2) {
			throw new IllegalArgumentException("BoundingBox must be an array of the length 2");
		}
		this.bbox = bbox;
	}

	/**
	 * Calculates the volume of the box
	 * 
	 * @return the volume of the box
	 */
	public double getVolume() {
		double length = getDepth();
		double width = getWidth();
		double height = getHeight();
		return height * width * length;
	}

	/**
	 * Calculates the center of the bounding box
	 * 
	 * @return the center of the bounding box
	 */
	public Vector3d getCenter() {
		return bbox[0].plus(bbox[1].minus(bbox[0]).mult(0.5));
	}

	/**
	 * Getter for the corner array
	 * 
	 * @return the array containing the corner points
	 */
	public Vector3d[] getBox() {
		return bbox;
	}

	/**
	 * Calculates the width of the bounding box
	 * 
	 * @return the width of the bounding box
	 */
	public double getWidth() {
		return bbox[1].getY() - bbox[0].getY();
	}

	/**
	 * Calculates the height of the bounding box
	 * 
	 * @return the height of the bounding box
	 */
	public double getHeight() {
		return bbox[1].getZ() - bbox[0].getZ();
	}

	/**
	 * Calculates the depth of the bounding box
	 * 
	 * @return the depth of the bounding box
	 */
	public double getDepth() {
		return bbox[1].getX() - bbox[0].getX();
	}

	/**
	 * Returns the length of the longest side of the bounding box, ignoring the
	 * height. Only X and Y axis are considered
	 * 
	 * @return the length of the longest side in X or Y direction
	 */
	public double getLongestSide() {
		double width = getWidth();
		double depth = getDepth();
		if (width > depth) {
			return width;
		} else {
			return depth;
		}
	}

	/**
	 * Calculates the direction vector from the lower corner to the upper corner
	 * 
	 * @return the direction vector from the lower corner to the upper corner
	 */
	public Vector3d getDiagonal() {
		return bbox[1].minus(bbox[0]);
	}

	/**
	 * Calculates the distance between the corner points
	 * 
	 * @return the distance between the corner points
	 */
	public double getDiagonalLength() {
		return bbox[1].getDistance(bbox[0]);
	}

	public static BoundingBox of(FeatureMapper mapper) {
		return BoundingBoxCalculator.calculateBoundingBox(mapper);
	}
	
	@Override
	public String toString() {
		return Arrays.toString(bbox);
	}

	public static BoundingBox of(List<Polygon> lod1Polygons, List<Polygon> lod2Polygons, List<Polygon> lod3Polygons,
			List<Polygon> lod4Polygons) {
		return BoundingBoxCalculator.calculateBoundingBox(lod1Polygons, lod2Polygons, lod3Polygons, lod4Polygons);

	}
}
